-----------------------------------------------------------------------------
--
-- Tag				:= { Statement } @
-- Statement		:= literal | "[" RValue "]"
-- RValue			:= Relation
-- Relation			:= Expression [ ("==" | "~=" | ">" | "<" | "<=" | ">=") Expression ]
-- Expression		:= Term [ ("+" | "-") Term ]
-- Factor			:= Chain [ ("*" | "/") Chain ]
-- Chain			:= Primary { ":" Primary }
-- Primary			:= ConstantNumber | ConstantString | Identifier ["(" ParameterList ")"] | ( RValue )
-- ParameterList	:= RValue {, RValue }
--
-----------------------------------------------------------------------------

AloftTag = { }
AloftTagMethods = { }
AloftTagDirectAccessData = { isBoss = true, level = true, name = true, healthBarValue = true, healthBarMaxValue = true }
AloftTagData = 
{ 
	healthBarValue 		= { events = "Aloft:OnHealthBarValueChanged" },
	healthBarMaxValue 	= { events = "Aloft:OnHealthBarValueChanged" },
}

-----------------------------------------------------------------------------

AloftTagMethods.angle = function(aloftData, value) return value and value ~= "" and ("<%s>"):format(value) end
AloftTagMethods.bracket = function(aloftData, value) return value and value ~= "" and ("(%s)"):format(value) end
AloftTagMethods.brace = function(aloftData, value) 	return value and value ~= "" and ("{%s}"):format(value) end

AloftTagMethods.color = function(aloftData, value, r, g, b)
	return value and ("|cff%02x%02x%02x%s|r"):format(r*255, g*255, b*255, value)
end

AloftTagMethods.colorIf = function(aloftData, value, condition, r, g, b)
	return value and condition and ("|cff%02x%02x%02x%s|r"):format(r*255, g*255, b*255, value) or value
end

AloftTagMethods.healthBarDeficit = function(aloftData, value)
	return value and (aloftData.healthBarMaxValue - aloftData.healthBarValue)
end
AloftTagData.healthBarDeficit = { events = "Aloft:OnHealthBarValueChanged" }

AloftTagMethods.healthPercent = function(aloftData, value)
	return value and floor(100*aloftData.healthBarValue / aloftData.healthBarMaxValue)
end
AloftTagData.healthPercent = { events = "Aloft:OnHealthBarValueChanged" }

AloftTagMethods.isGroup = function(aloftData, value)
	return aloftData.healthBarMaxValue ~= 100 and value
end

AloftTagMethods.isNotGroup = function(aloftData, value)
	return aloftData.healthBarMaxValue == 100 and value
end

AloftTagMethods.hexColor = function(aloftData, value, colorValue)
	return value and ("|cff%s%s|r"):format(colorValue, value or "")
end

AloftTagMethods.hexColorIf = function(aloftData, value, condition, colorValue)
	return value and condition and ("|cff%s%s|r"):format(colorValue, value) or value
end

AloftTagMethods.hideIf = function(aloftData, value, condition)
	return not condition and value
end

AloftTagMethods.hideZero = function(aloftData, value)
	return value ~= 0 and value
end

AloftTagMethods.isFriendly = function(aloftData, value)
	return (aloftData.type == "friendlyPlayer" or aloftData.type == "friendlyNPC") and value
end

AloftTagMethods.isFriendlyNPC = function(aloftData, value)
	return aloftData.type == "friendlyPlayer" and value
end

AloftTagMethods.isFriendlyPlayer = function(aloftData, value)
	return aloftData.type == "friendlyPlayer" and value
end

AloftTagMethods.isHostile = function(aloftData, value)
	return aloftData.type == "hostile" and value
end

AloftTagMethods.isNeutral = function(aloftData, value)
	return aloftData.type == "neutral" and value
end

AloftTagMethods.isMaxHealth = function(aloftData, value)
	return aloftData.healthBarValue == aloftData.healthBarMaxValue and value
end
AloftTagData.isMaxHealth = { events = "Aloft:OnHealthBarValueChanged" }

AloftTagMethods.lower = function(aloftData, value)
	if type(value) == "string" then
		return value:lower()
	end
	return value
end

AloftTagMethods.negate = function(aloftData, value)
	if type(value) == "number" then
		return -value
	end
	return value
end

AloftTagMethods["not"] = function(aloftData, value, condition)
	return not condition and value
end

AloftTagMethods.percent = function(aloftData, value)
	if type(value) == "number" then
		return value.."%"
	end
	return value
end

AloftTagMethods.prefix = function(aloftData, value, prefix)
	return value and prefix..value
end

AloftTagMethods.rgb = function(aloftData, value, r, g, b)
	return value and ("|cff%02x%02x%02x%s|r"):format(r, g, b, value)
end

AloftTagMethods.rgbIf = function(aloftData, value, condition, r, g, b)
	return value and condition and ("|cff%02x%02x%02x%s|r"):format(r, g, b, value) or value
end

AloftTagMethods.short = function(aloftData, value)
	if type(value) == "number" then
		-- Use a 'binary' tree to determine our range, rather than the old sequential search
		-- This gives us maximum 3 comparisons
		if value <= -10000 then
			if value <= -100000 then
				if value <= -1000000 then
					return ("%.2fm"):format(value / 1000000)
				else
					return ("%dk"):format(value / 1000)
				end
			else
				return ("%.1fk"):format(value / 1000)
			end
		else
			if value < 100000 then
				if value < 10000 then
					return ("%d"):format(value)
				else
					return ("%.1fk"):format(value / 1000)
				end
			else
				if value < 1000000 then
					return ("%dk"):format(value / 1000)
				else
					return ("%.2fm"):format(value / 1000000)
				end				
			end
		end
	end
	return value
end

AloftTagMethods.suffix = function(aloftData, value, suffix)
	return value and value..suffix
end

AloftTagMethods.upper = function(aloftData, value)
	if type(value) == "string" then
		return value:upper()
	end
	return value
end

AloftTagMethods.white 	= function(aloftData, value) return value and ("|r%s"):format(value) end
AloftTagMethods.red 	= [[hexColor("ff0000")]]
AloftTagMethods.green 	= [[hexColor("00ff00")]]
AloftTagMethods.blue 	= [[hexColor("0000ff")]]
AloftTagMethods.cyan 	= [[hexColor("00ffff")]]
AloftTagMethods.fuchsia = [[hexColor("ff00ff")]]
AloftTagMethods.yellow 	= [[hexColor("ffff00")]]
AloftTagMethods.gray 	= [[hexColor("afafaf")]]
AloftTagMethods.difficultyColor = function(aloftData, value) return value and ("|cff%02x%02x%02x%s|r"):format(aloftData.levelTextR*255, aloftData.levelTextG*255, aloftData.levelTextB*255, value) end

-----------------------------------------------------------------------------

local parseTag
local parseTagOffset
local tagMethods = { }
local overrideEvents = { }
local numeric = false

local function GetByte()
	while parseTag:byte(parseTagOffset) == 32 do 
		parseTagOffset = parseTagOffset + 1
	end 
	return parseTag:byte(parseTagOffset) or 0
end

local function AdvanceByte()
	parseTagOffset = parseTagOffset + 1
end

-----------------------------------------------------------------------------

local function TagErrorStringMethod() return "!!Tag Error!!" end
local function TagError(x) Aloft:Print(x) end

local function EmptyTable(table) for k in pairs(table) do table[k] = nil end end

function AloftTag:CoerceToNumber(text)
	if not text or text == "" then return "0" end
	if numeric then return text end
	return ("(%s or 0)"):format(text)
end

function AloftTag:ParseParameterList(dataTable)
	if GetByte() == 40 then -- byte == '('
		AdvanceByte()
		local byte = GetByte()
		if byte == 41 then 
			AdvanceByte()
			return ""
		end

		local parameterListString = (", %s"):format(self:ParseRValue(dataTable))
		
		while parseTagOffset <= #parseTag and GetByte() == 44 do -- byte == ','
			AdvanceByte()
			parameterListString = ("%s, %s"):format(parameterListString, self:ParseRValue(dataTable))
		end

		if GetByte() == 41 then -- byte == ')'
			AdvanceByte()
		else
			dataTable.error = true
			TagError("Missing end of parameter list \")\"")
		end
		
		return parameterListString
	else
		return ""
	end
end

function AloftTag:ParsePrimary(dataTable, functionString)
	-- Either an integer, a string or an LValue (value or function call)
	numeric = false
	
	local byte = GetByte()
	if byte >= 48 and byte <= 59 then -- byte >= '0' && byte <= '9'
		local startOffset = parseTagOffset
		local endOffset = startOffset
		local decimalPointFound = false
		while true do
			endOffset = endOffset + 1
			byte = parseTag:byte(endOffset) or 0
			if byte == 46 then -- '.'
				if decimalPointFound then
					break
				else
					decimalPointFound = true
				end
			elseif byte < 48 or byte > 59 then
				break
			end
		end
		parseTagOffset = endOffset
		if functionString then
			return ("(%s and %s)"):format(functionString, parseTag:sub(startOffset, endOffset-1))
		else
			numeric = true
			return parseTag:sub(startOffset, endOffset-1)
		end
	elseif byte == 34 then	-- byte == '\"'
		local startOffset = parseTagOffset+1
		local endOffset = startOffset + 1
		while true do
			if endOffset > #parseTag then
				TagError("Missing closing quote")
				dataTable.error = true
				return ""
			end
			byte = parseTag:byte(endOffset) or 0
			if byte == 34 then -- byte == '\"'
				break
			elseif byte == 92 then	-- byte == '\\'
				endOffset = endOffset + 2
			else
				endOffset = endOffset + 1
			end
		end
		parseTagOffset = endOffset+1
		if functionString then
			return ("%s and %q"):format(functionString, parseTag:sub(startOffset, endOffset-1))
		else
			return ("%q"):format(parseTag:sub(startOffset, endOffset-1))
		end
		
	elseif (byte >= 65 and byte <= 90) or (byte >= 97 and byte <= 122) then	-- byte >= 'A' && byte <= 'Z' || byte >= 'a' && byte <= 'z'
		local startOffset = parseTagOffset
		local endOffset = startOffset
		while (byte >= 65 and byte <= 90) or (byte >= 97 and byte <= 122) do
			endOffset = endOffset + 1
			byte = parseTag:byte(endOffset) or 0
		end
		parseTagOffset = endOffset
		local identifier = parseTag:sub(startOffset, endOffset-1)
		local data = AloftTagData[identifier]
		if data then
            if data.events then
				if type(data.events) == "string" then
                    dataTable.events[data.events] = true
				elseif type(data.events) == "table" then
					for _,v in pairs(data.events) do
						dataTable.events[v] = true
					end
                end
            end
            if data.overrideEvents then
                if type(data.overrideEvents) == "string" then
                    overrideEvents[data.overrideEvents] = true
				elseif type(data.overrideEvents) == "table" then
					for _,v in pairs(data.overrideEvents) do
						overrideEvents[v] = true
					end
                end
            end
            if data.data then
                if type(data.data) == "string" then
                    dataTable.data[data.data] = true
				elseif type(data.data) == "table" then
					for _,v in pairs(data.data) do
						dataTable.data[v] = true
					end
                end
            end
		end
		if AloftTagDirectAccessData[identifier] then
			if functionString then
				return functionString.." and aloftData."..identifier
			else
				return "aloftData."..identifier
			end
		elseif AloftTagMethods[identifier] then
			if type(AloftTagMethods[identifier]) == "string" then
				local tag, offset = parseTag, parseTagOffset
				parseTag, parseTagOffset = AloftTagMethods[identifier], 1
				local returnValue = self:ParsePrimary(dataTable, functionString)
				parseTag, parseTagOffset = tag, offset
				return returnValue
			else
				tagMethods[identifier] = true	-- this marks the method as required, and it's inserted as an upvalue
				local parameterList = self:ParseParameterList(dataTable)
				return ("tagMethod%s(aloftData, %s%s)"):format(identifier, functionString or "true", parameterList)
			end
		else
			TagError("Invalid tag: "..identifier)
			dataTable.error = true
			return ""
		end
	elseif (byte == 40) then -- '('
		AdvanceByte()
		local result = self:ParseRValue(dataTable)
		local byte = GetByte()
		if byte == 41 then
			AdvanceByte()
			return ("(%s)"):format(result)
		else
			TagError("Missing \")\"")
			dataTable.error = true
			return ""
		end
	end
end

function AloftTag:ParseChain(dataTable)
	local functionString = self:ParsePrimary(dataTable)
	while GetByte() == 58 do	-- while tag[offset] == ':'
		AdvanceByte()
		functionString = self:ParsePrimary(dataTable, functionString)
	end
	return functionString
end

function AloftTag:ParseFactor(dataTable)
	local result = self:ParseChain(dataTable)
	local byte = GetByte()
	while byte == 42 or byte == 47 do	-- '*' or '/'
		AdvanceByte()
		result = self:CoerceToNumber(result)
		result = ("%s %s %s"):format(result, string.char(byte), self:CoerceToNumber(self:ParseFactor(dataTable)))
		numeric = true
		byte = GetByte()
	end
	return result
end

function AloftTag:ParseExpression(dataTable)
	local result = self:ParseFactor(dataTable)
	local byte = GetByte()
	while byte == 43 or byte == 45 do	-- '+' or '-'
		AdvanceByte()
		result = self:CoerceToNumber(result)
		result = ("%s %s %s"):format(result, string.char(byte), self:CoerceToNumber(self:ParseFactor(dataTable)))
		numeric = true
		byte = GetByte()
	end
	return result
end

function AloftTag:ParseRelation(dataTable)
	local result = self:ParseExpression(dataTable)

	local byte = GetByte()
	local byte2 = parseTag:byte(parseTagOffset+1)
	if byte == 61 then	-- '='
		AdvanceByte()
		if byte2 == 61 then
			AdvanceByte()
			return result.."=="..self:ParseExpression(dataTable)
		else
			TagError("Unexpected \"=\"")
			dataTable.error = true
			return ""
		end
	elseif byte == 126 then -- '~'
		AdvanceByte()
		if byte2 == 61 then
			AdvanceByte()
			return result.."~="..self:ParseExpression(dataTable)
		else
			TagError("Unexpected \"=\"")
			dataTable.error = true
			return ""
		end
	elseif byte == 60 then -- '<'
		AdvanceByte()
		result = self:CoerceToNumber(result)
		if byte2 == 61 then
			AdvanceByte()
			return ("%s <= %s"):format(result, self:CoerceToNumber(self:ParseExpression(dataTable)))
		else
			return ("%s < %s"):format(result, self:CoerceToNumber(self:ParseExpression(dataTable)))
		end
	elseif byte == 62 then -- '>'
		AdvanceByte()
		result = self:CoerceToNumber(result)
		if byte2 == 61 then
			AdvanceByte()
			return ("%s >= %s"):format(result, self:CoerceToNumber(self:ParseExpression(dataTable)))
		else
			return ("%s > %s"):format(result, self:CoerceToNumber(self:ParseExpression(dataTable)))
		end
	else
		return result
	end
end

function AloftTag:ParseRValue(dataTable)
	return self:ParseRelation(dataTable)
end

-- Initialise the compile cache with an empty tag, something which would otherwise need some extra special-cased code
local compileCache =
{
	[""] = { data = { }, events = { }, tag = "", method = function() return "" end },
}

local headerTable = { }
local functionStringList = { }
function AloftTag:CreateTag(tag)
	parseTag = tag or ""

	if compileCache[parseTag] then
		return compileCache[parseTag]
	end

	parseTagOffset = 1
	local dataTable = { data = { }, events = { }, tag = parseTag }

	table.insert(functionStringList, "") -- it's replaced with all the locals required later
	table.insert(functionStringList, "return function(aloftData) return ")

	while parseTagOffset <= #parseTag and not dataTable.error do
		if parseTagOffset > 1 then
			table.insert(functionStringList, "..")
		end
		if parseTag:byte(parseTagOffset) == 91 then	-- if tag[offset] == '['
			AdvanceByte()
			table.insert(functionStringList, ([[(%s or "")]]):format(self:ParseRValue(dataTable) or ""))
			if GetByte() == 93 then -- if tag[offset] == ']'
				AdvanceByte()
			elseif not dataTable.error then
				TagError("Tag not closed - missing \"]\"")
				dataTable.error = true
			end
		else
			-- literal section
			local startOffset = parseTagOffset
			local endOffset = startOffset+1
			while endOffset <= #parseTag and parseTag:byte(endOffset) ~= 91 do
				endOffset = endOffset + 1
			end
			table.insert(functionStringList, ("%q"):format(parseTag:sub(startOffset, endOffset-1)))
			parseTagOffset = endOffset
		end
	end

	for k in pairs(overrideEvents) do
		dataTable.events[k] = nil
		overrideEvents[k] = nil
	end

	if dataTable.error then
		dataTable.method = TagErrorStringMethod
		EmptyTable(dataTable.data)
		EmptyTable(dataTable.events)
	else
		for k in pairs(tagMethods) do
			table.insert(headerTable, ("local tagMethod%s = AloftTagMethods[\"%s\"]\n"):format(k, k))
		end
		functionStringList[1] = table.concat(headerTable)
		EmptyTable(headerTable)
		table.insert(functionStringList, " end")
		local functionString = table.concat(functionStringList)
--[[
		Aloft:Debug(string.format("Tag: %s", parseTag))
		for event in pairs(dataTable.events) do
			Aloft:Debug(string.format(" - event: %s", event))
		end
		for data in pairs(dataTable.data) do
			Aloft:Debug(string.format(" - data: %s", data))
		end
		Aloft:Debug("--- Compiled method ---\n"..functionString)
--]]--
		dataTable.method = loadstring(functionString)()
		compileCache[parseTag] = dataTable
	end

	EmptyTable(tagMethods)
	EmptyTable(functionStringList)

	return dataTable
end

-----------------------------------------------------------------------------
